import pygame

from pygame.locals import *

from OpenGL.GL import *

from tdgl import part,hox,animator,objpart,picking
from math import cos, sin, atan2, radians, degrees
import random

import ai


def make_stats(**kwd):
    d = dict(hp=0,ep=0,power=0,moves=0,attacks=0,
                jump=0,speed=0,capacity=0,reach=0,
                clever=0)
    d.update(kwd)
    return d

class Robot(part.Group):
    """Robot (group of robot parts acting in unison)"""
    _default_geom = {"pos":(0,0,0),"angle":0}
    def __init__(self,name,pieces,team,*args,**kwd):
        super(Robot,self).__init__(name,pieces,*args,**kwd)
        self.team = team
        self.name = name
        self.anim = animator.Mutator(self._geom)
        self.hp = 1
        self.ep = 0
        self.power = 0
        self.moves = 0
        self.attacks = 0
        self.jump = 0.0
        self.speed = 0
        self.capacity = 0
        self.reach = 0
        self.clever = 0
        self.calc_stats()
        self.health = self.hp
        self.program = []
    def calc_stats(self):
        self.hp = 0
        self.ep = 0
        self.power = 0
        self.moves = 0
        self.attacks = 0
        self.jump = 0.0
        self.speed = 0
        self.capacity = 0
        self.reach = 0
        self.clever = 0
        for p in self.contents:
            self.hp += p.stats["hp"]
            self.ep += p.stats["ep"]
            self.power += p.stats["power"]
            self.moves += p.stats["moves"]
            self.attacks += p.stats["attacks"]
            self.jump += p.stats["jump"]
            self.speed += p.stats["speed"]
            self.capacity += p.stats["capacity"]
            self.reach += p.stats["reach"]
            self.clever += p.stats["clever"]

    def step(self,ms):
        self.anim.step(ms)
        super(Robot,self).step(ms)

    def scramble(self):
        self.program = random.shuffle(self.program)

    def next_action(self):
        if self.program:
            return self.program[0]

    def done_action(self):
        if self.program:
            print "Done:",self.program.pop(0)

    def replace_action(self,actions):
        if self.program:
            self.program[:1] = actions

    def animate_turn(self,angle,ms):
        newangle = self.angle + angle
        self.anim.change('angle',newangle,ms)

    def animate_move(self,tox,toy,toz,ms):
        x,y,z = self.pos
        if z > toz + 0.001:
            seq = [(x,y,z),(tox,toy,z),(tox,toy,toz)]
            mss = [ms//2, ms//2]
        elif z < toz - 0.001:
            seq = [(x,y,z),(x,y,toz),(tox,toy,toz)]
            mss = [ms//2, ms//2]
        else:
            seq = [(x,y,z),(tox,toy,toz)]
            mss = [ms]
        self.anim.sequence('pos',seq,mss)
        leg = self["legs"]
        leg.animate_move(ms)
    def animate_attack(self,ms,attacktype="Attack"):
        if attacktype == "AttackL":
            arm = self["arml"]
        else:
            arm = self["armr"]
        if arm:
            arm.animate_attack(self,ms)

    def render(self,mode):
        if mode == "PICK":
            picking.label(("Robot",self))
        super(Robot,self).render(mode)
        if mode == "PICK":
            picking.nolabel()

def direction(angle):
    # dx, dy as nearest NSEW direction to angle
    dxf = cos(radians(angle))
    dyf = sin(radians(angle))
    # round to nearest int
    return int(round(dxf)), int(round(dyf))

    
class RoboFight:
    """Controller for robot battles"""
    def __init__(self,stackmap,robots,brb=None):
        self.stackmap = stackmap
        self.robots = robots
        self.step_for_state = {
            "Waiting":self.step_while_waiting,
            "Fighting":self.step_while_fighting,
            "Victory":self.step_while_waiting,
            "Defeat":self.step_while_waiting,
            "Title":self.step_while_in_title
            }
        self.brb = brb
        self.ai_thread = None
        self.change_state("Waiting")
        self.fight_step_time = 750 #ms
        self.robot_moves = {}
        pygame.mixer.music.load("data/music/waiting.ogg")
        pygame.mixer.music.play(-1)
        for r in robots:
            x,y,z = r.pos
            self.robot_moves[int(round(x)),int(round(y))] = r

    def change_state(self,newstate):
        pygame.mixer.music.fadeout(1000)
        assert newstate in self.step_for_state
        self.state = newstate
        self.wait(0)
        if self.brb:
            self.brb.on = (newstate == "Fighting")
        if newstate == "Fighting":
            self.stop_thinking()
        else:
            self.start_thinking()
        if self.state == "Fighting":
            pygame.mixer.music.load("data/music/fighting.ogg")
            pygame.mixer.music.play(-1)
        elif self.state == "Waiting":
            pygame.mixer.music.load("data/music/waiting.ogg")
            pygame.mixer.music.play(-1)
        elif self.state == "Title":
            pygame.mixer.music.load("data/music/title.ogg")
            pygame.mixer.music.play(-1)
            

    def start_thinking(self):
        self.stop_thinking()
        self.ai_thread = ai.RobotAIThread(self.stackmap,self.robots,"enemy")
        self.ai_thread.setDaemon(True)
        self.ai_thread.start()
    def stop_thinking(self):
        if self.ai_thread and self.ai_thread.isAlive():
            self.ai_thread.interrupt()
            self.ai_thread.join(3.0)
        self.ai_thread = None

    def wait(self,ms):
        self.counter = 0
        self.when_next = ms

    def step(self,ms):
        self.counter += ms
        if self.counter >= self.when_next:
            self.step_for_state[self.state](ms)

    def step_while_waiting(self,ms):
        self.counter = 0

    def step_while_fighting(self,ms):
        robots_list = sorted(
            self.robots,
            key=lambda r: -(r.speed + r.clever))
        for r in robots_list:
            action = r.next_action()
            if action is None:
                continue
            x,y,z = r.pos
            x,y = int(round(x)), int(round(y))
            dx,dy = direction(r.angle)

            # robot has something to do
            if action == "Walk":
                h = self.stackmap.floor_height(x+dx,y+dy)
                if (h is not None and
                    not self.robot_at(x+dx,y+dy) and
                    abs(h-z) <= r.jump):
                    r.animate_move(x+dx,y+dy,h,self.fight_step_time)
                    self.robot_moves[x,y] = None
                    self.robot_moves[x+dx,y+dy] = r
                r.done_action()
            elif action == "Left":
                r.animate_turn(90,self.fight_step_time)
                r.done_action()
            elif action == "Right":
                r.animate_turn(-90,self.fight_step_time)
                r.done_action()
            elif action.startswith("Attack"):
                r2 = self.robot_at(x+dx,y+dy)
                h = self.stackmap.floor_height(x+dx,y+dy)
                if h is None or abs(h - z) > r.reach:
                    # cannot attack because of height
                    r.done_action()
                elif r2 is None:
                    r.animate_attack(self.fight_step_time,action)
                    r.done_action()
                elif r2.team == r.team and r.clever > 0:
                    # Don't hit own team
                    r.done_action()
                else: # Attack!
                    r.animate_attack(self.fight_step_time,action)
                    r.done_action()
                    self.resolve_attack(r,r2,action)
            else:
                # unknown action, e.g. "dance"
                doit = getattr(r,"animate_" + action,None)
                if doit:
                    doit(self.fight_step_time)
                else:
                    print "Robot %s tried to %s!" % (r._name,action)
                r.done_action()
        # See who's still alive and still moving
        robots_still_to_move = False
        allies = 0
        enemies = 0
        for r in robots_list:
            if r.health > 0:
                if r.team == "enemy":
                    enemies += 1
                elif r.team == "ally":
                    allies += 1
                if r.next_action():
                    robots_still_to_move = True
            else:
                self.remove_robot(r)
        # if all are completely moved, end the turn
        if robots_still_to_move:
            self.wait(self.fight_step_time)
        else:
            if allies == 0 and self.state != "Title":
                self.change_state("Defeat")
            elif enemies == 0:
                self.change_state("Victory")
            else:
                self.change_state("Waiting")

    def step_while_in_title(self,ms):
        self.step_while_fighting(ms)
        if self.state == "Waiting":
            self.start_thinking()
            self.change_state("Title")
    def remove_robot(self,r):
        self.robots.remove(r)
        for xy, r2 in self.robot_moves.items():
            if r2 == r:
                self.robot_moves[xy] = None
                break
        r._expired = True
        
    def robot_at(self,x,y):
        # robot known to be in or moving into a square
        return self.robot_moves.get((x,y))

    def resolve_attack(self,attacker,defender,attack_type):
        """ decide how much damage and of what kind
        TODO: make this cleverer
        """
        # height advantage
        attack_power = attacker.power + attacker.pos[2] - defender.pos[2]
        defender.health -= attack_power
        arm = attacker[{"Attack":"armr","AttackL":"arml"}[attack_type]]
        if arm:
            arm.special_attack(attacker,defender)

class RobotPiece(part.Part):
    _default_style={
        'obj-filename':'robot-head-scope.obj',
        'obj-pieces':()}
    _default_geom={
        'pos':(0.0,0.0,0.0),
        'angle':0.0,
        'Xangle':0.0,
        'Yangle':0.0}
    _style_attributes=('obj-filename','obj-pieces')
    def __init__(self,name,*args,**kwd):
        super(RobotPiece,self).__init__(name,*args,**kwd)
        self.anim = animator.Mutator(self._geom)
        self.hox = []
        self.obj = None
        self.prepare()
    def setup_style(self):
        glPushAttrib(GL_LIGHTING_BIT)
        glEnable(GL_LIGHTING)
        glEnable(GL_CULL_FACE)
    def setdown_style(self):
        glPopAttrib()
    def step(self,ms):
        self.anim.step(ms)
    def render(self,mode):
        glCullFace(GL_BACK)
        hox.process(self.hox,self.getgeom,self.drawpiece)
    def drawpiece(self,pname):
        if self.obj:
            self.obj.drawpiece(pname)
    def prepare(self):
        fname = self.getstyle("obj-filename")
        if fname:
            self.obj = objpart.get_obj(fname)
            pieces = self.getstyle('obj-pieces')
            if not pieces:
                pieces = self.obj.pieces()
            self.hox = [(hox.DRAW,p)for p in pieces ]
    def animate_move(self,ms):
        pass
    def special_attack(self,attacker,defender):
        pass

class ScopeHead(RobotPiece):
    _default_style = {"obj-filename":"robot-head-scope.obj"}
    _default_geom = {"pos":(0.0,0.0,0.55)}
    def __init__(self,name,*args,**kwd):
        super(ScopeHead,self).__init__(name,*args,**kwd)
        self.anim = animator.Mutator(self._geom)
        self.stats = make_stats(hp=10,ep=20,moves=7,
                                attacks=1,clever=2)
        self.hox = []
        self.obj = None
        self.prepare()

class ScopePlusHead(RobotPiece):
    _default_style = {"obj-filename":"robot-head-scope.obj"}
    _default_geom = {"pos":(0.0,0.0,0.55)}
    def __init__(self,name,*args,**kwd):
        super(ScopePlusHead,self).__init__(name,*args,**kwd)
        self.anim = animator.Mutator(self._geom)
        self.stats = make_stats(hp=30,ep=20,moves=8,
                                attacks=2,clever=2)
        self.hox = []
        self.obj = None
        self.prepare()

class VandalHead(RobotPiece):
    _default_style = {"obj-filename":"robot-head-vandal.obj"}
    _default_geom = {"pos":(0.0,0.0,0.55)}
    def __init__(self,name,*args,**kwd):
        super(VandalHead,self).__init__(name,*args,**kwd)
        self.anim = animator.Mutator(self._geom)
        self.stats = make_stats(hp=20,ep=20,moves=8,
                                attacks=2,clever=2)
        self.hox = []
        self.obj = None
        self.prepare()

class CyclopseHead(RobotPiece):
    _default_style = {"obj-filename":"robot-head-cyclopse.obj"}
    _default_geom = {"pos":(0.0,0.0,0.55)}
    def __init__(self,name,*args,**kwd):
        super(CyclopseHead,self).__init__(name,*args,**kwd)
        self.stats = make_stats(hp=10,ep=25,moves=5,
                                attacks=1,clever=1)

class AndroidHead(RobotPiece):
    _default_style = {"obj-filename":"robot-head-android.obj"}
    _default_geom = {"pos":(0.0,0.0,0.55)}
    def __init__(self,name,*args,**kwd):
        super(AndroidHead,self).__init__(name,*args,**kwd)
        self.stats = make_stats(hp=10,ep=45,moves=6,
                                attacks=2,clever=1)

class MilitaryHead(RobotPiece):
    _default_style = {"obj-filename":"robot-head-military.obj"}
    _default_geom = {"pos":(0.0,0.0,0.55)}
    def __init__(self,name,*args,**kwd):
        super(MilitaryHead,self).__init__(name,*args,**kwd)
        self.stats = make_stats(hp=20,ep=45,moves=3,
                                attacks=1,clever=1)

class BushidoHead(RobotPiece):
    _default_style = {"obj-filename":"robot-head-bushido.obj"}
    _default_geom = {"pos":(0.0,0.0,0.55)}
    def __init__(self,name,*args,**kwd):
        super(BushidoHead,self).__init__(name,*args,**kwd)
        self.stats = make_stats(hp=20,ep=45,moves=6,
                                attacks=3,clever=1)

class GorillaHead(RobotPiece):
    _default_style = {"obj-filename":"robot-head-gorilla.obj"}
    _default_geom = {"pos":(0.08,0.0,0.45)}
    def __init__(self,name,*args,**kwd):
        super(GorillaHead,self).__init__(name,*args,**kwd)
        self.stats = make_stats(hp=25,ep=80,moves=7,
                                attacks=4,clever=5)

class SpringBody(RobotPiece):
    _default_style = {"obj-filename":"robot-body-spring.obj"}
    _default_geom = {"pos":(0.0,0.0,0.55)}
    def __init__(self,name,*args,**kwd):
        super(SpringBody,self).__init__(name,*args,**kwd)
        self.stats = make_stats(hp=50,power=8,capacity=4,
                                reach=1)

class GlassBody(RobotPiece):
    _default_style = {"obj-filename":"robot-body-glass.obj"}
    _default_geom = {"pos":(0.0,0.0,0.55)}
    def __init__(self,name,*args,**kwd):
        super(GlassBody,self).__init__(name,*args,**kwd)
        self.stats = make_stats(hp=65,power=6,capacity=3,
                                reach=1)

class AndroidBody(RobotPiece):
    _default_style = {"obj-filename":"robot-body-android.obj"}
    _default_geom = {"pos":(0.0,0.0,0.55)}
    def __init__(self,name,*args,**kwd):
        super(AndroidBody,self).__init__(name,*args,**kwd)
        self.anim = animator.Mutator(self._geom)
        self.stats = make_stats(hp=45,power=4,capacity=3,
                                reach=1)

class MilitaryBody(RobotPiece):
    _default_style = {"obj-filename":"robot-body-military.obj"}
    _default_geom = {"pos":(0.0,0.0,0.55)}
    def __init__(self,name,*args,**kwd):
        super(MilitaryBody,self).__init__(name,*args,**kwd)
        self.anim = animator.Mutator(self._geom)
        self.stats = make_stats(hp=75,power=8,capacity=3,
                                reach=1)

class BushidoBody(RobotPiece):
    _default_style = {"obj-filename":"robot-body-bushido.obj"}
    _default_geom = {"pos":(0.0,0.0,0.55)}
    def __init__(self,name,*args,**kwd):
        super(BushidoBody,self).__init__(name,*args,**kwd)
        self.anim = animator.Mutator(self._geom)
        self.stats = make_stats(hp=60,power=12,capacity=3,
                                reach=1)

class VandalBody(RobotPiece):
    _default_style = {"obj-filename":"robot-body-vandal.obj"}
    _default_geom = {"pos":(0.0,0.0,0.55)}
    def __init__(self,name,*args,**kwd):
        super(VandalBody,self).__init__(name,*args,**kwd)
        self.stats = make_stats(hp=150,power=10,capacity=4,
                                reach=1)

class GorillaBody(RobotPiece):
    _default_style = {"obj-filename":"robot-body-gorilla.obj"}
    _default_geom = {"pos":(0.0,0.0,0.55)}
    def __init__(self,name,*args,**kwd):
        super(GorillaBody,self).__init__(name,*args,**kwd)
        self.stats = make_stats(hp=200,power=12,capacity=6,
                                reach=1)


class StupidLegs(RobotPiece):
    _default_style = {"obj-filename":"robot-legs-stupid.obj"}
    _default_geom = {"pos":(0.0,0.0,0.15),"leg-angle":0.0}
    def __init__(self,name,*args,**kwd):
        super(StupidLegs,self).__init__(name,*args,**kwd)
        self.stats = make_stats(moves=1,jump=1.0,speed=2)
        self.hox = [[(hox.ROT,"leg-angle",hox.Y,1.0),
                     (hox.DRAW,"LegR")],
                    [(hox.ROT,"leg-angle",hox.Y,-1.0),
                     (hox.DRAW,"LegL")],
                    (hox.DRAW,"Pelvis"),
                    (hox.DRAW,"Torso"),]
    def animate_move(self,ms):
        print "move"
        self.anim.sequence("leg-angle",[0,-30,30,0],[ms/4,ms/2,ms/4])

class VandalLegs(RobotPiece):
    _default_style = {"obj-filename":"robot-legs-vandal.obj"}
    _default_geom = {"pos":(0.0,0.0,0.15),"leg-angle":0.0}
    def __init__(self,name,*args,**kwd):
        super(VandalLegs,self).__init__(name,*args,**kwd)
        self.stats = make_stats(moves=1,jump=2.0,speed=2)
        self.hox = [[(hox.ROT,"leg-angle",hox.Y,1.0),
                     (hox.DRAW,"LegR")],
                    [(hox.ROT,"leg-angle",hox.Y,-1.0),
                     (hox.DRAW,"LegL")],
                    (hox.DRAW,"Pelvis"),
                    (hox.DRAW,"Torso"),]
    def animate_move(self,ms):
        print "move"
        self.anim.sequence("leg-angle",[0,-30,30,0],[ms/4,ms/2,ms/4])
        

class GorillaLegs(RobotPiece):
    _default_style = {"obj-filename":"robot-legs-gorilla.obj"}
    _default_geom = {"pos":(0.0,0.0,0.35),"leg-angle":0.0}
    def __init__(self,name,*args,**kwd):
        super(GorillaLegs,self).__init__(name,*args,**kwd)
        self.stats = make_stats(moves=1,jump=2.5,speed=2)
        self.hox = [[(hox.ROT,"leg-angle",hox.Y,1.0),
                     (hox.DRAW,"LegR")],
                    [(hox.ROT,"leg-angle",hox.Y,-1.0),
                     (hox.DRAW,"LegL")],]
    def animate_move(self,ms):
        print "move"
        self.anim.sequence("leg-angle",[0,-50,50,0],[ms/4,ms/2,ms/4])


class HoverLegs(RobotPiece):
    _default_style = {"obj-filename":"robot-legs-hoverball.obj"}
    _default_geom = {"pos":(0.0,0.0,0.15)}
    def __init__(self,name,*args,**kwd):
        super(HoverLegs,self).__init__(name,*args,**kwd)
        self.stats = make_stats(moves=1,jump=0.75,speed=1)


class AndroidLegs(RobotPiece):
    _default_style = {"obj-filename":"robot-legs-android.obj"}
    _default_geom = {"pos":(0.0,0.0,0.15)}
    def __init__(self,name,*args,**kwd):
        super(AndroidLegs,self).__init__(name,*args,**kwd)
        self.stats = make_stats(moves=1,jump=1.5,speed=3)

class MilitaryLegs(RobotPiece):
    _default_style = {"obj-filename":"robot-legs-tracks.obj"}
    _default_geom = {"pos":(0.0,0.0,0.15)}
    def __init__(self,name,*args,**kwd):
        super(MilitaryLegs,self).__init__(name,*args,**kwd)
        self.stats = make_stats(moves=10,jump=0.25,speed=6)

class BushidoLegs(RobotPiece):
    _default_style = {"obj-filename":"robot-legs-bushido.obj"}
    _default_geom = {"pos":(0.0,0.0,0.15)}
    def __init__(self,name,*args,**kwd):
        super(BushidoLegs,self).__init__(name,*args,**kwd)
        self.stats = make_stats(moves=3,jump=3.0,speed=4)

class NofaceLegs(RobotPiece):
    _default_style = {"obj-filename":"robot-legs-bushido.obj"}
    _default_geom = {"pos":(0.0,0.0,0.15)}
    def __init__(self,name,*args,**kwd):
        super(NofaceLegs,self).__init__(name,*args,**kwd)
        self.stats = make_stats(moves=8,attacks=2,jump=3.0,speed=4)

class ChainsawArm(RobotPiece):
    _default_style = {"obj-filename":"robot-arm-chainsaw.obj"}
    _default_geom = {"pos":(0.0,-0.2,0.45),"lift-angle":0.0}
    """ Kickass --Badger"""
    def __init__(self,name,flip=False,*args,**kwd):
        super(ChainsawArm,self).__init__(name,*args,**kwd)
        self.stats = make_stats(ep=10,power=5)
        if flip:
            self.setgeom("pos",(0.0,0.2,0.45))
        self.hox = [(hox.ROT,"lift-angle",hox.Y,-1.0)]+self.hox
        self.sound = pygame.mixer.Sound("data/sfx/chainsaw.ogg")
    def animate_attack(self,body,ms):
        self.sound.play()
        self.anim.sequence("lift-angle",[0,180,0],[0,ms])
        x,y,z = body.pos
        dx,dy = direction(body.angle)
        body.anim.sequence("pos",[(x,y,z),(x+(dx*0.6),y+(dy*0.6),z),(x,y,z)],[ms/2,ms/2])


class AxeArm(RobotPiece):
    _default_style = {"obj-filename":"robot-arm-axe.obj"}
    _default_geom = {"pos":(0.0,-0.2,0.45),"lift-angle":0.0}
    def __init__(self,name,flip=False,*args,**kwd):
        super(AxeArm,self).__init__(name,*args,**kwd)
        self.stats = make_stats(ep=10,power=2)
        if flip:
            self.setgeom("pos",(0.0,0.2,0.45))
        self.hox = [(hox.ROT,"lift-angle",hox.Y,-1.0)]+self.hox
        self.sound = pygame.mixer.Sound("data/sfx/blade.ogg")
    def animate_attack(self,body,ms):
        self.sound.play()
        self.anim.sequence("lift-angle",[0,180,0],[ms/4,ms*0.75])
        x,y,z = body.pos
        dx,dy = direction(body.angle)
        body.anim.sequence("pos",[(x,y,z),(x+(dx*0.6),y+(dy*0.6),z),(x,y,z)],[ms/2,ms/2])
        print "Attack!"

class HellstarArm(RobotPiece):
    _default_style = {"obj-filename":"robot-arm-hellstar.obj"}
    _default_geom = {"pos":(0.0,-0.2,0.45),"lift-angle":0.0}
    def __init__(self,name,flip=False,*args,**kwd):
        super(HellstarArm,self).__init__(name,*args,**kwd)
        self.stats = make_stats(ep=10,power=8)
        if flip:
            self.setgeom("pos",(0.0,0.2,0.45))
        self.hox = [(hox.ROT,"lift-angle",hox.Y,1.0)]+self.hox
        self.sound = pygame.mixer.Sound("data/sfx/impact.ogg")
    def animate_attack(self,body,ms):
        self.sound.play()
        self.anim.sequence("lift-angle",[0,360,0],[ms,0])
        x,y,z = body.pos
        dx,dy = direction(body.angle)
        body.anim.sequence("pos",[(x,y,z),(x+(dx*0.6),y+(dy*0.6),z),(x,y,z)],[ms/2,ms/2])


class LanceArm(RobotPiece):
    _default_style = {"obj-filename":"robot-arm-lance.obj"}
    _default_geom = {"pos":(0.0,-0.2,0.45),"lift-angle":0.0,"thrust":0.0}
    def __init__(self,name,flip=False,*args,**kwd):
        super(LanceArm,self).__init__(name,*args,**kwd)
        self.stats = make_stats(ep=10,power=4,reach=1)
        if flip:
            self.setgeom("pos",(0.0,0.2,0.45))
        self.hox = [(hox.ROT,"lift-angle",hox.Y,-1.0),
                    (hox.DRAW,"Shoulder"),
                    (hox.DRAW,"Arm"),
                    (hox.MOVE,"thrust",hox.Z,1.0),
                    (hox.DRAW,"blade")
                    ]
        self.sound = pygame.mixer.Sound("data/sfx/blade.ogg")
    def animate_attack(self,body,ms):
        self.sound.play()
        self.anim.sequence("lift-angle",[0,90,90,0],[ms*0.1,ms*0.8,ms*0.1])
        self.anim.sequence("thrust",[0.0,-0.2,0],[ms*0.4,ms*0.4])

class TenderiserArm(RobotPiece):
    _default_style = {"obj-filename":"robot-arm-tenderiser.obj"}
    _default_geom = {"pos":(0.0,-0.2,0.45),"lift-angle":0.0}
    def __init__(self,name,flip=False,*args,**kwd):
        super(TenderiserArm,self).__init__(name,*args,**kwd)
        self.stats = make_stats(ep=10,power=8)
        if flip:
            self.setgeom("pos",(0.0,0.2,0.45))
        self.hox = [(hox.ROT,"lift-angle",hox.Y,-1.0)]+self.hox
        self.sound = pygame.mixer.Sound("data/sfx/impact.ogg")
    def animate_attack(self,body,ms):
        self.sound.play()
        self.anim.sequence("lift-angle",[0,180,90,180,90,180,0],[ms*0.15,ms*0.1,ms*0.2,ms*0.1,ms*0.2,ms*0.15])
        x,y,z = body.pos
        dx,dy = direction(body.angle)
        body.anim.sequence("pos",[(x,y,z),(x+(dx*0.6),y+(dy*0.6),z),(x,y,z)],[ms/2,ms/2])

class KatanaArm(RobotPiece):
    _default_style = {"obj-filename":"robot-arm-katana.obj"}
    _default_geom = {"pos":(0.0,-0.2,0.45),"lift-angle":0.0}
    def __init__(self,name,flip=False,*args,**kwd):
        super(KatanaArm,self).__init__(name,*args,**kwd)
        self.stats = make_stats(ep=10,power=11)
        if flip:
            self.setgeom("pos",(0.0,0.2,0.45))
        self.hox = [(hox.ROT,"lift-angle",hox.Y,-1.0)]+self.hox
        self.sound = pygame.mixer.Sound("data/sfx/blade.ogg")
    def animate_attack(self,body,ms):
        self.sound.play()
        self.anim.sequence("lift-angle",[0,180,0],[ms/4,ms*0.75])
        x,y,z = body.pos
        dx,dy = direction(body.angle)
        body.anim.sequence("pos",[(x,y,z),(x+(dx*0.6),y+(dy*0.6),z),(x,y,z)],[ms/2,ms/2])
        print "Attack!"

class MaulArm(RobotPiece):
    _default_style = {"obj-filename":"robot-arm-maul.obj"}
    _default_geom = {"pos":(0.0,-0.2,0.45),"lift-angle":0.0}
    def __init__(self,name,flip=False,*args,**kwd):
        super(MaulArm,self).__init__(name,*args,**kwd)
        self.stats = make_stats(ep=10,power=10)
        if flip:
            self.setgeom("pos",(0.0,0.2,0.45))
        self.hox = [(hox.ROT,"lift-angle",hox.Y,-1.0)]+self.hox
        self.sound = pygame.mixer.Sound("data/sfx/impact.ogg")
    def animate_attack(self,body,ms):
        self.sound.play()
        self.anim.sequence("lift-angle",[0,180,0],[ms/4,ms*0.75])
        x,y,z = body.pos
        dx,dy = direction(body.angle)
        body.anim.sequence("pos",[(x,y,z),(x+(dx*0.6),y+(dy*0.6),z),(x,y,z)],[ms/2,ms/2])
        print "Attack!"

class VandalPart(RobotPiece):
    _default_style = {"obj-filename":"robot-arm-vandalpart.obj"}
    _default_geom = {"pos":(0.0,-0.2,0.45),"lift-angle":0.0}
    def __init__(self,name,flip=False,*args,**kwd):
        super(VandalPart,self).__init__(name,*args,**kwd)
        self.stats = make_stats(ep=10,power=18,attacks=1)
        if flip:
            self.setgeom("pos",(0.0,0.2,0.45))
        self.hox = [(hox.ROT,"lift-angle",hox.Y,-1.0)]+self.hox
        self.sound = pygame.mixer.Sound("data/sfx/impact.ogg")
    def animate_attack(self,body,ms):
        self.sound.play()
        self.anim.sequence("lift-angle",[0,180,145,180,0],[ms*0.3,ms*0.2,ms*0.2,ms*0.3])
        x,y,z = body.pos
        dx,dy = direction(body.angle)
        body.anim.sequence("pos",[(x,y,z),(x+(dx*0.6),y+(dy*0.6),z),(x,y,z)],[ms/2,ms/2])
        print "Attack!"

class GorillaArm(RobotPiece):
    _default_style = {"obj-filename":"robot-arm-gorilla.obj"}
    _default_geom = {"pos":(0.0,-0.3,0.7),"lift-angle":0.0}
    def __init__(self,name,flip=False,*args,**kwd):
        super(GorillaArm,self).__init__(name,*args,**kwd)
        self.stats = make_stats(ep=20,power=10)
        if flip:
            self.setgeom("pos",(0.0,0.3,0.7))
        self.hox = [(hox.ROT,"lift-angle",hox.Y,-1.0)]+self.hox
        self.sound = pygame.mixer.Sound("data/sfx/impact.ogg")
    def animate_attack(self,body,ms):
        self.sound.play()
        self.anim.sequence("lift-angle",[0,180,0],[0,ms])
        x,y,z = body.pos
        dx,dy = direction(body.angle)
        body.anim.sequence("pos",[(x,y,z),(x+(dx*0.6),y+(dy*0.6),z),(x,y,z)],[ms/2,ms/2])
        print "Attack!"

class DrainArm(RobotPiece):
    _default_style = {"obj-filename":"robot-arm-drain.obj"}
    _default_geom = {"pos":(0.0,-0.3,0.7),"lift-angle":0.0,"spin":0.0}
    def __init__(self,name,flip=False,*args,**kwd):
        super(DrainArm,self).__init__(name,*args,**kwd)
        self.stats = make_stats(ep=20,power=8)
        if flip:
            self.setgeom("pos",(0.0,0.3,0.7))
        self.hox = [(hox.ROT,"lift-angle",hox.Y,-1.0),
                    (hox.DRAW,"Arm"),
                    (hox.DRAW,"Wrist"),
                    (hox.ROT,"spin",hox.Z,1.0),
                    (hox.DRAW,"Blades")]
        self.sound = pygame.mixer.Sound("data/sfx/chainsaw.ogg")
    def animate_attack(self,body,ms):
        self.sound.play()
        self.anim.sequence("lift-angle",[0,90,90,0],[ms*0.1,ms*0.8,ms*0.1])
        self.anim.sequence("spin",[0,720,0],[ms,0])
        x,y,z = body.pos
        dx,dy = direction(body.angle)
        body.anim.sequence("pos",[(x,y,z),(x+(dx*0.6),y+(dy*0.6),z),(x,y,z)],[ms/2,ms/2])
        print "Attack!"
    def special_attack(self,attacker,defender):
        attacker.health += 5
        defender.program = []
