"""backdrop.py

A Backdrop is responsible for clearing the colour and depth buffers,
and may also draw something onto the screen as a background.

Best to put a Backdrop inside an viewgroup.OrthoGroup so the projection is
correct.

Copyright May 2005 by Peter Harris

Released under the terms of the GNU General Public License
see www.gnu.org for details.
"""

from OpenGL import GL
import part, texture, lighting, stylesheet

BUFFERBITS = GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT

class Backdrop(part.Part):
    _default_style={'background-colour':(0,0,0,1)}
    _style_attributes=('background-colour',False)    # yuk!
    _active = False # don't keep a group active if just a background in it
    def __init__(self,name='',**kwd):
        super(Backdrop,self).__init__(name,**kwd)
        self.restyle()
    def setup_geom(self):
        pass
    def setdown_geom(self):
        pass
    def setup_style(self):
        GL.glClearColor(*self._style['background-colour'])
    def render(self,mode):
        if mode == "OPAQUE":   # transparency or picking pass irrelevant
            GL.glClear(BUFFERBITS)

class TextureBackdrop(Backdrop):
    _default_geom = {'distance':100.0}
    _default_style = {'background-colour':(0,0,0,1), 'colour':(1,1,1,1)}
    _style_attributes = ('background-colour','colour','texture')
    _active = False
    tex = None
    
    def __init__(self,name='',**kwd):
        super(TextureBackdrop,self).__init__(name,**kwd)

    def prepare(self):
        texname = self._style.get('texture')
        if texname:
            self.tex = texture.get_texture(texname)
            self._has_transparent = self.tex.transparent
        else:
            self.tex = None
            self._has_transparent = False
        self.colour = self._style['colour']
        self._has_transparent = self._has_transparent or  (self.colour[3] < 1.0)

    def setup_style(self):
        """Enable textures and bind self.tex texture object, if it has one."""
        tex = self.tex
        colour = self.colour
        if tex:
            texture.enable()
            tex.bind()
            if self._has_transparent:
                GL.glEnable(GL.GL_BLEND)
                GL.glClearColor(*self._style['background-colour'])
            else:
                GL.glDisable(GL.GL_BLEND)
        lighting.disable()
        GL.glColor(*colour)
    
    def render(self,mode):
        if mode == "OPAQUE":
            GL.glClear(BUFFERBITS)  # only clear on first pass
        elif mode == "PICK":
            return  # no need to pick
        GL.glDepthMask(GL.GL_FALSE)
        v = GL.glVertex3f
        t = GL.glTexCoord2f
        z = -self._geom['distance']
        GL.glBegin(GL.GL_QUADS)
        t(0,0) ; v(-1,-1,z)
        t(1,0) ; v( 1,-1,z)
        t(1,1) ; v( 1, 1,z)
        t(0,1) ; v(-1, 1,z)
        GL.glEnd()
        if mode != 'TRANSPARENT':
            GL.glDepthMask(GL.GL_TRUE)
