"""TextPanel is a rectangular area with lines of text on it.

Copyright Sep 2007 by Peter Harris
Released under the terms of the GNU General Public License v3 or later

"""
from OpenGL.GL import *
import pygame
import part, texturefont, texture, lighting, stylesheet, usualGL, picking

__all__ = ('TextPanel',)

got_multitexture = None

class TextPanel(part.Part):
    """A panel containing zero or more lines of text, with an optional
    background colour and outline.
    It provides array-like methods to make it easier to change the lines of text.
    If a selected line is specified, it will be highlighted with the
    highlight foreground and background colours.
    If there are more lines than the maxlines argument, a slice of the lines
    will be shown that includes the selected line.
    """
    _default_geom = {'pos':(0,0,0)}
    _default_style ={
        'colour':(1,1,1,1),
        'background-colour':(0,0,0.5,0.5),
        'border-colour':(1,1,1,1), 'border-width':2,
        'border-radius':0, 'border-round':0,
        'highlight-radius':0, 'highlight-round':0,
        'highlight-colour':(1,1,0,1),
        'highlight-background':(0,0,1,1), 'pitch':20,
        'texture':None, 'texture-repeat':None,
        'multitexture':None,
        'padding':0.5,
        'highlight-padding':0.1,
        'fontname':None}
    _style_attributes = _default_style.keys()
    _has_transparent = True # usually see-through
    
    def __init__(self,name,lines=[],**kwd):
        super(TextPanel,self).__init__(name,**kwd)
        self.all_lines = lines[:]
        self.maxlines = kwd.get('maxlines')
        self.selected = kwd.get('selected')
        self.firstline = 0

    def __getitem__(self,i):
        return self.all_lines[i]

    def __setitem__(self,i,s):
        self.all_lines[i] = str(s)
        self.prepare()

    def __delitem__(self,i):
        del self.all_lines[i]
        self.prepare()

    def __len__(self):
        return len(self.all_lines)

    def append(self,s):
        self.all_lines.append(str(s))
        self.prepare()

    def prepare(self):
        """Prepare to display some lines of text"""
        numlines = len(self.all_lines)
        if self.selected is None:
            pass
        elif self.selected >= numlines:
            self.selected = numlines - 1
        elif self.selected < 0:
            self.selected = 0
        if self.maxlines is not None:
            numlines = min(self.maxlines, numlines)
            if self.selected is not None:
                if self.firstline + numlines <= self.selected:
                    self.firstline = self.selected - numlines + 1
                elif self.selected < self.firstline:
                    self.firstline = self.selected
        else:
            self.firstline = 0
        
        style = self._style
        self.font = texturefont.get_font(style.get('fontname'))
        self.height = numlines
        self.width = w = max([self.font.width(s) for s in self.all_lines])
        pad = style['padding']
        hpad = style['highlight-padding']
        self.borderpoints = stylesheet.border_points(
            w + pad, self.height + pad,
            style['border-radius'], style['border-round'])
        self.highlightpoints = stylesheet.border_points(
            w + hpad, 1.0 + hpad, style['highlight-radius'], style['highlight-round'])
        texname = style.get('texture')
        if texname:
            self.tex = texture.get_texture(texname)
            self.texpoints = self.borderpoints
            rep = style.get('texture-repeat')
            if rep == 'scale':
                self.texpoints = [(x/self.width + 0.5, y/self.height + 0.5)
                                    for x,y in self.borderpoints]
            tex_transparent = self.tex.transparent
        else:
            self.tex = None
            tex_transparent = False
            self.texpoints = []

        global got_multitexture
        if got_multitexture is None:
            got_multitexture = "GL_ARB_multitexture" in glGetString(GL_EXTENSIONS)
        mtexname = style.get('multitexture')
        if got_multitexture and mtexname:
            self.mtex = texture.get_texture(mtexname)
        else:
            self.mtex = None

    def event(self, event):
        """On a keypress, move highlight bar and possibly change index of
        first line"""
        def scroll(dy):
            numLines = len(self.all_lines)
            if self.selected is not None:
                self.selected = (self.selected + dy) % numLines
                self.prepare()
                return True
            elif self.maxlines is not None and self.maxlines < numLines:
                self.firstline += dy
                self.firstline = min(numLines - self.height -1, self.firstline)
                self.firstline = max(0, self.firstline)
                self.prepare()
                return True
            else:
                return None
        if event.type != pygame.KEYDOWN:
            return None
        if event.key == pygame.K_RETURN and self.selected is not None:
            return (self._name,self.selected, self.all_lines[self.selected])
        elif event.key == pygame.K_UP:
            return scroll(-1)
        elif event.key == pygame.K_DOWN:
            return scroll(1)
        else:
            return None

    def setup_geom(self):
        """Translate to position, scale according to text pitch"""
        scale = 2.0 / self._style.get('pitch',2.0)
        glPushMatrix()
        glLoadIdentity()
        glTranslate(*self.pos)
        glScalef(scale,scale,scale)

    def render(self,mode='OPAQUE'):
        """Draw the outline, background and text.
        Because of various style changes, no separate
        setup_style() method would be much use.
        """
        if mode == 'PICK':
            picking.label((self,None))
        lighting.disable()
        style = self._style
        bgcolour = style.get('background-colour')
        bdcolour = style.get('border-colour')
        bdwidth = style.get('border-width')
        if bgcolour is not None:
            self.draw_background(mode,bgcolour)
        if bdcolour is not None and bdwidth:
            self.draw_outline(mode,bdcolour,bdwidth)
        self.draw_text(mode)
        if mode == 'PICK':
            picking.nolabel()

    def draw_outline(self,mode,colour,linewidth):
        """Draw outline using a line loop"""
        texture.disable()
        transparent = (colour[3] < 1.0)
        if mode != 'PICK':
            glColor(*colour)
        glLineWidth(linewidth)
        glEnable(GL_LINE_SMOOTH)
        glBegin(GL_LINE_STRIP)
        v = glVertex2f
        for x,y in self.borderpoints:
            v(x,y)
        glEnd()

    def draw_background(self,mode,colour):
        """Draw filled in background using a triangle fan"""
        if not self.borderpoints:
            return
        if mode != 'PICK':
            glColor(*colour)
            if self.tex and self.texpoints:
                self.tex.bind()
                texture.enable()
        v = glVertex3f
        tc = glTexCoord2f
        glBegin(GL_TRIANGLE_FAN)
        tc(0.5,0.5)
        v(0,0,-0.02)
        if self.texpoints:
            for (x,y),(s,t) in zip(self.borderpoints, self.texpoints):
                tc(s,t)
                v(x,y,-0.02)
        else:
            for x,y in self.borderpoints:
                v(x,y,-0.02)
        glEnd()

    def draw_highlight(self,mode):
        """Draw coloured highlight bar"""
        if mode == 'PICK':
            return  # not needed
        hbg = self._style['highlight-background']
        hlz = -0.015
        w = self.width * 0.5
        glColor(*hbg)
        v = glVertex3f
        texture.disable()
        glBegin(GL_TRIANGLE_FAN)
        v(w,0.5,hlz)
        for x,y in self.highlightpoints:
            v(x + w,y + 0.5,hlz)
        glEnd()

    def draw_text(self,mode):
        """Draw each line of text, starting from top left"""
        fg = self._style['colour']
        glTranslate(self.width * -0.5, (self.height * 0.5) - 1.0, 0.0)
        highlight_on = self.selected is not None
        line0 = self.firstline
        if highlight_on:
            hfg = self._style['highlight-colour']
            lineSel = self.selected - line0 # which visible line is highlighted
        else:
            if self.mtex:
                glActiveTexture(GL_TEXTURE1)
                self.mtex.bind()
                glEnable(GL_TEXTURE_GEN_S)
                glEnable(GL_TEXTURE_GEN_T)
                glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR)
                glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR)
                glTexGenfv(GL_S, GL_OBJECT_PLANE, (1,0,0,0))
                glTexGenfv(GL_T, GL_OBJECT_PLANE, (0,1,0,0))
                glEnable(GL_TEXTURE_2D)
                glActiveTexture(GL_TEXTURE0)
            glColor(*fg)
        for i, line in enumerate(self.all_lines[line0:line0 + self.height]):
            if highlight_on:
                if i == lineSel:
                    self.draw_highlight(mode)
                    glColor(*hfg)
                else:
                    glColor(*fg)
            if mode == 'PICK':
                picking.label((self,i))
            self.font.render(line,stay=True)
            glTranslate(0.0,-1.0,0.0)
        if self.mtex:
            glActiveTexture(GL_TEXTURE1)
            glDisable(GL_TEXTURE_2D)
            glActiveTexture(GL_TEXTURE0)
            
