""" TextureFont is a range of display lists corresponding to textured quads,
    constructed from a ttf font using pygame.font.

    Copyright 2007 Peter Harris, released under the terms of the GNU General Public License
    (v3 or later).

"""
from __future__ import division
from weakref import WeakValueDictionary

import pygame
from pygame import font, image
from OpenGL import GL,GLU
from OpenGL.GL import GL_TEXTURE_2D, GL_RGBA, GL_UNSIGNED_BYTE, GL_COMPILE
from OpenGL.GL import GL_QUADS, GL_ALPHA_TEST, GL_GREATER, GL_INTENSITY

from math import log

DEFAULT_INTERNAL_FORMAT = GL.GL_INTENSITY4

__all__ = ('TextureFont','get_font')

try:    # use builtin set type if available
    set
except NameError:
    from sets import Set as set

def is_a_power_of_2(x):
    """Binary trick: x & -x yields lowest set bit in x.
    For powers of 2 it's the only bit"""
    return x == (x & -x)

def next_power_of_2(n):
    """Smallest power of 2 >= n"""
    if is_a_power_of_2(n):
        return n
    else:
        return 2 ** (int(log(n,2)) + 1)

class TextureFont:
    def __init__(self,fontfile=None,size=64,chars=128,fromchar=32,debug=False):
        self.nchars = chars
        self.offset = GL.glGenLists(self.nchars)
        self.defined_chars = set()
        self.textures = {}
        self.widths = {}
        self.fromchar = fromchar
        self.debug = debug
        try:
            getFont = font.Font
            fnt = getFont(fontfile,size)
        except IOError:
            getFont = font.SysFont
            fnt = getFont(fontfile,size)
        theight = next_power_of_2(size)
        self.theight = theight  # texture height
        # guess actual font size to make it fit in texture neatly
        # e.g. if font height is 50 and texture size is 64, try increasing
        # size by 63/50
        newsize = int(size * (theight - 1) / fnt.get_height())
        if size != newsize:
            size = newsize
            fnt = getFont(fontfile,size)  # throw one away. Oh well
        self.fnt = fnt
        self.fnt_height = fnt.get_height()
    def define_char(self,char):
        """Define a single character as a display list to a texture"""
        width,height = self.fnt.size(unichr(char))
        # get power-of-2 width and height, make an empty transparent image
        twidth = next_power_of_2(width)
        theight = self.theight
        if self.debug:
            print "Defining char", char, (width,height), (twidth,theight)
        bare_data = "\x00\x00\x00\x00" * (twidth * theight)
        tsurface = image.fromstring(bare_data,(twidth, theight),"RGBA")
        # make image of character and blit into the surface
        try:
            csurface = self.fnt.render(unichr(char),False,(255,255,255,255))
        except pygame.error:
            csurface = self.fnt.render(unichr(char),True,(255,255,255,255))
        tsurface.blit(csurface,(0,0))
        # make a texture
        img_data = image.tostring(tsurface,"RGBA")
        tname = GL.glGenTextures(1)
        GL.glEnable(GL_TEXTURE_2D)
        GL.glBindTexture(GL_TEXTURE_2D,tname)
        GLU.gluBuild2DMipmaps(
            GL_TEXTURE_2D, DEFAULT_INTERNAL_FORMAT, twidth, theight,
            GL_RGBA, GL_UNSIGNED_BYTE,
            img_data)
        self.textures[char] = tname # keep
        # build display list of a quad covered by the part of the texture containing the char
        s = 0.0
        v = GL.glVertex2f
        t = GL.glTexCoord2f
        ds = 1.0 / twidth   # s-coord increment for one texel
        dx = 1.0 / theight  # x-coord increment for one texel at scale=1.0
        GL.glNewList(char+self.offset,GL_COMPILE)
        GL.glBindTexture(GL_TEXTURE_2D,tname)
        sw = width * ds
        xw = width * dx
        GL.glBegin(GL_QUADS)
        t(s, 1.0)
        v(0.0, 0.0)
        t(s+sw, 1.0)
        v(xw, 0.0)
        t(s+sw, 0.0)
        v(xw, 1.0)
        t(s, 0.0)
        v(0.0, 1.0)
        GL.glEnd()
        GL.glTranslatef(xw,0.0,0.0)  # move right
        GL.glEndList()
        self.widths[char] = xw
        self.defined_chars.add(unichr(char))

    def define_undefined_chars(self,s):
        undefined = set(s) - self.defined_chars
        for c in undefined:
            n = ord(c)
            if n >= self.fromchar and n < (self.fromchar + self.nchars):
                self.define_char(n)
    def __del__(self):
        """delete any lists allocated"""
        try:
            GL.glDeleteLists(self.offset,self.nchars)
            GL.glDeleteTextures(self.textures.values())
        except:
            pass
    def render(self,string,stay=True):
        """Render a string at (0,0,0), chars are about 1.0 high. Translates
        modelview coordinates if stay is False."""
        if not string:
            return
        self.define_undefined_chars(string)
        GL.glListBase(self.offset)
        GL.glEnable(GL_TEXTURE_2D)
        GL.glEnable(GL_ALPHA_TEST)
        GL.glAlphaFunc(GL_GREATER,0.1)
        if stay:
            GL.glPushMatrix()
        GL.glCallLists([ord(c) for c in string])
        if stay:
            GL.glPopMatrix()
    def width(self,string):
        self.define_undefined_chars(string)
        return sum([self.widths[ord(c)] for c in string])
    def bind(self,char=32):
        """Select the texture used for a char of the font, mostly for debugging purposes"""
        GL.glBindTexture(GL_TEXTURE_2D,self.textures[char])

font_pool = WeakValueDictionary()

def get_font(filename,**kwds):
    """Get a font by filename, re-use one already built if possible"""
    if filename not in font_pool:
        fnt = TextureFont(filename,**kwds)
        font_pool[filename] = fnt
    return font_pool[filename]

if __name__ == '__main__':
    import sys
    import pygame
    pygame.init()
    if sys.argv[1:]:
        fontfile = sys.argv[1]
    else:
        fontfile = None
    tf = TextureFont(fontfile,debug=True)
